import numpy as np
from math import *

#ms_per_step=2

#def step_ms(x):
#    return x*ms_per_step

def SpikeTrain_ISI(x):
    result = []
    last = -1
    for i, x in enumerate(x):
        if x > 0:
            if last!=-1:
                result.append(i-last)
            last=i

    return result

def get_bins(number, xrightlim):
    #xrightlim = max_neuron_frequency()
    binwidth = xrightlim / number
    return np.arange(0, xrightlim, binwidth)

def get_distribution_markers(data):
    mean = np.mean(data)
    std = np.std(data)
    return mean-std, mean, mean+std

def plot_histogram(axis, data, bins, mean=True, std=True):
    axis.hist(data, bins=bins)
    for x,c in zip(get_distribution_markers(data), ['green', 'red', 'green']):
        axis.axvline(x=x, color=c)


def plot_3D_histogram(axis, data, bins, time_steps, mean=True, std=True, x_label_caption='Spike Rate Hz'):

    for i in range(time_steps):
        z = i * 10

        step = int(len(data) / time_steps)
        ys = data[i * step:(i + 1) * step]
        hist, bins = np.histogram(ys, bins=bins)
        xs = (bins[:-1] + bins[1:]) / 2
        width = bins[1] - bins[0]

        c = np.random.rand(3)
        axis.bar(xs, hist, color=c, ec=c, alpha=0.8, width=width, zdir='y', zs=z)

    axis.set_xlabel(x_label_caption)
    axis.set_zlabel('Nr')
    axis.set_ylabel('Recording Blocks')

    axis.set_yticklabels([])

#def plot_3D_hist(act):
#    fig = plt.figure()
#    ax = fig.add_subplot(111, projection='3d')
#    plot_3D_histogram(ax, act, 40, 10)
#    plt.show()



def get_reconstruction_activations(origin_activations, input_grid_width, input_grid_height):
    dim = int(floor(sqrt(len(origin_activations))))
    result_pic_width = (dim + 1) * (input_grid_width + 1)
    result_pic_height = (dim + 1) * (input_grid_height + 1)

    # print(dim)
    # print(result_pic_width)
    # print(result_pic_height)

    border_value = np.max(origin_activations)

    x = 0
    y = 0

    result = np.ones((result_pic_height, result_pic_width)) * border_value

    # print(origin_activations)

    for activations in origin_activations:
        inp_syn = activations[0:input_grid_width * input_grid_height].reshape((input_grid_height, input_grid_width))
        result[y * (input_grid_height + 1):y * (input_grid_height + 1) + input_grid_height, x * (input_grid_width + 1):x * (input_grid_width + 1) + input_grid_width] = inp_syn
        x += 1
        if x > dim:
            y += 1
            x = 0

    return result



#def plot_reconstruction_activations(origin_activations, input_grid_width, input_grid_height, ax=None):
#    result=get_reconstruction_activations(origin_activations, input_grid_width, input_grid_height)

    # colormap = matplotlib.cm.gist_gray
#    colormap = plt.cm.gist_gray

    # colormap.set_bad(color='red')
    # result = np.ma.array(result, mask=(result == np.array(0)))

#    if ax is not None:
#        ax.imshow(result, cmap=colormap, interpolation='nearest')
#    else:
#        plt.imshow(result, cmap=colormap, interpolation='nearest')


# def get_whole_Network_reconstruction_image(self, neuron_dst_group, input_source, x_elements, y_elements, element_rows=-1, element_cols=-1, add_activity=True, include_GABA=True):
#    if element_rows == -1: element_rows = input_source.width
#    if element_cols == -1: element_cols = input_source.height
#
#    reconstructions=self.get_reconstruction(, source_group_neuron_indexes, input_source, synapse_weight_attribute_name, recostruction_steps)
#
#    return self.get_combined_RGB_Image(elements_g=weights, elements_r=inh_synapses, element_rows=element_rows, element_cols=element_cols, x_elements=x_elements, y_elements=y_elements)

# rec=network.get_reconstruction(Cortex_PC_Neurons, list(range(10)), LGN_PC_Neurons, 'GLU_Synapses', 1)#{'GLU_Synapses':1,'GABA_Synapses':-1}, just_nuerons=[] just_synapses=[]
# network.plot_reconstruction_activations(rec, 10, 10)
# plt.show()


def dif_width_vstack(matrixes, reshapes, max_width):
    #print(reshapes, max_width)

    for i in range(len(matrixes)):
        if reshapes[i][0] < max_width:
            t = []
            for e in range(len(matrixes[i])):
                t.append(np.hstack([matrixes[i][e].reshape(reshapes[i][0], reshapes[i][1]), np.zeros((max_width-reshapes[i][0], reshapes[i][1]))]).flatten())
            matrixes[i] = np.array(t)

    return np.hstack(matrixes)

def upscale(img,factor):
    h = img.shape[0]
    w = img.shape[1]
    r = img[:,:,0].repeat(factor).reshape((h, w*factor)).repeat(factor, axis=0).reshape((h*factor, w*factor))
    g = img[:,:,1].repeat(factor).reshape((h, w*factor)).repeat(factor, axis=0).reshape((h*factor, w*factor))
    b = img[:,:,2].repeat(factor).reshape((h, w*factor)).repeat(factor, axis=0).reshape((h*factor, w*factor))
    return np.stack([r, g, b],axis=-1)

def get_whole_Network_weight_image(neuron_dst_group, plot_inh_synapses=False, neuron_src_groups=None, add_activity=False, individual_norm=False, exc_weight_attr='GLU_Synapses', inh_weight_attr='GABA_Synapses', activations=None):

    exc_synapses = []
    inh_synapses = []

    shapes = []

    temp_inh_syns = neuron_dst_group.afferent_synapses['GABA']

    for syn in neuron_dst_group.afferent_synapses['GLU']:
        if neuron_src_groups is None or syn.src in neuron_src_groups:

            found = None
            for inh_syn in temp_inh_syns:
                if syn.src == inh_syn.src and syn.dst == inh_syn.dst:
                    found = inh_syn

            if found is not None:
                inh_synapses.append(getattr(found, inh_weight_attr).copy())
                #inh_synapses.append(found.GABA_Synapses.copy())
            else:
                inh_synapses.append(np.zeros(getattr(syn, exc_weight_attr).shape))
                #inh_synapses.append(np.zeros(syn.GLU_Synapses.shape))

            #print(syn.src.width, syn.src.height)

            exc_synapses.append(getattr(syn, exc_weight_attr).copy())
            shapes.append(np.array([syn.src.width, syn.src.height*syn.src.depth]))

    shapes = np.array(shapes)

    #print(shapes, shapes[:, 0])
    max_cols = np.max(shapes[:, 0])
    row_sum = np.sum(shapes[:, 1])

    weights = dif_width_vstack(exc_synapses, shapes, max_cols)
    inh_weights = dif_width_vstack(inh_synapses, shapes, max_cols)

    #print(weights.shape, inh_weights.shape, row_sum, max_cols)

    if individual_norm:
        weights = weights/np.max(weights, axis=1)[:, None]

    #print(neuron_dst_group.depth)

    return get_RGB_neuron_weight_image([inh_weights, weights], weight_dim=[row_sum, max_cols], neuron_dim=[neuron_dst_group.height*neuron_dst_group.depth, neuron_dst_group.width], activities=activations)
    #return get_combined_RGB_Image(g_weight_mat=weights, r_weight_mat=inh_weights, element_rows=max_cols, element_cols=row_sum, x_elements=neuron_dst_group.width, y_elements=neuron_dst_group.height*neuron_dst_group.depth)


def get_group_block_mat_reconstruction_image(network, source_block, synapse_weight_attribute_name, recostruction_steps, individual_norm=False, print_status=False):

    if type(source_block) == np.ndarray:
        source_block=list(source_block.flatten())
    elif not type(source_block) == list:
        source_block=[source_block]

    #reconstructions=
    #for neuron_group in source_block:


    #image = network.get_reconstruction(group, None, LGN_PC_Neurons, 'GLU_Synapses', 3, individual_norm=True)
    # axes[i].imshow(get_RGB_neuron_weight_image([image, None, None], [group.height*group.depth, group.width], [LGN_PC_Neurons.height*LGN_PC_Neurons.depth, LGN_PC_Neurons.width], v_split_first=v_split))#pattern_f




def get_RGB_neuron_weight_image_pattern(rgb_weights, neuron_dim, pattern):
    return get_RGB_neuron_weight_image(rgb_weights, neuron_dim, pattern.get_pattern_dimension(), pattern.reconstruct_pattern)


def get_RGB_neuron_weight_image(rgb_weights, neuron_dim, weight_dim, weight_transform_func=None, v_split_first=False, activities=None):

    if weight_transform_func is None:
        temp=weight_dim.copy()
        weight_transform_func = lambda mat: mat.reshape(temp[0], temp[1])
    #else:


    if v_split_first:
        weight_dim[0]=int(weight_dim[0]/2)

    for w in rgb_weights:
        if w is not None:
            scale = 1
            w_max = np.max(w)
            if w_max > 0:
                scale = 1 / w_max
            w *= scale

    result_height = neuron_dim[0] * weight_dim[0] + neuron_dim[0] + 1
    result_width = neuron_dim[1] * weight_dim[1] + neuron_dim[1] + 1

    result = np.ones((result_height, result_width, 3))

    for y in range(neuron_dim[0]):
        for x in range(neuron_dim[1]):

            element_id = x + y * neuron_dim[0]

            ymin = (1 + y) + y * weight_dim[0]
            ymax = (1 + y) + (y + 1) * weight_dim[0]
            xmin = (1 + x) + x * weight_dim[1]
            xmax = (1 + x) + (x + 1) * weight_dim[1]

            result[ymin:ymax, xmin:xmax, :] *= 0

            if v_split_first:
                result[ymin:ymax, xmin:xmax, 0:2] = vsplit(weight_transform_func(rgb_weights[0][element_id]))
            else:
                for i, w in enumerate(rgb_weights):
                    if w is not None:
                        transformed = weight_transform_func(w[element_id])
                        result[ymin:ymax, xmin:xmax, i] = transformed

            if activities is not None:
                result[ymin:ymax, xmin:xmax, 2] += activities[element_id]

    return result

def vsplit(weight_mat):
    result = np.zeros((int(weight_mat.shape[0]/2), weight_mat.shape[1], 2))
    result[:, :, 0]=weight_mat[0:int(weight_mat.shape[0] / 2), :]
    result[:, :, 1]=weight_mat[int(weight_mat.shape[0] / 2):, :]
    return result

#def v_split_top(weight_mat):
#    return weight_mat[0:int(weight_mat.shape[0] / 2), :]

#def v_split_bottom(weight_mat):
#    return weight_mat[int(weight_mat.shape[1] / 2):, :]

#reconstruct_pattern
'''
def get_combined_RGB_Image(r_weight_mat=None, g_weight_mat=None, x_elements=0, y_elements=0, element_rows=0, element_cols=0, activities=None, h_split_rmat=False, transform_mat_func=None):

    if h_split_rmat:
        #r_weight_mat = v_split_top(r_weight_mat)
        #g_weight_mat = v_split_bottom(r_weight_mat)
        element_cols = int(element_cols / 2)

    if transform_mat_func is None:
        print('tmf')
        def default_transform_mat_func(mat):
            return mat.reshape(element_cols, element_rows)
        transform_mat_func = default_transform_mat_func
    else:
        dim = transform_mat_func(None)
        element_cols = dim[0]
        element_rows = dim[1]


    result_width = x_elements * element_rows + x_elements + 1
    result_height = y_elements * element_cols + y_elements + 1

    result = np.ones((result_height, result_width, 3))

    if r_weight_mat is not None:
        r_scale = 1
        r_max = np.max(r_weight_mat)
        if r_max > 0:
            r_scale = 1 / r_max

    if g_weight_mat is not None:
        g_scale = 1
        g_max = np.max(g_weight_mat)
        if g_max > 0:
            g_scale = 1 / g_max

    for y in range(y_elements):
        for x in range(x_elements):

            element_id = x + y * x_elements

            xmin = (1 + x) + x * element_rows
            xmax = (1 + x) + (x + 1) * element_rows
            ymin = (1 + y) + y * element_cols
            ymax = (1 + y) + (y + 1) * element_cols

            if r_weight_mat is not None:
                if h_split_rmat:
                    result[ymin:ymax, xmin:xmax, 0] = v_split_top(transform_mat_func(r_weight_mat[element_id])) * r_scale
                else:
                    result[ymin:ymax, xmin:xmax, 0] = transform_mat_func(r_weight_mat[element_id]) * r_scale
            else:
                result[ymin:ymax, xmin:xmax, 0] *= 0

            if g_weight_mat is not None:
                if h_split_rmat:
                    result[ymin:ymax, xmin:xmax, 0] = v_split_bottom(transform_mat_func(r_weight_mat[element_id])) * r_scale
                else:
                    result[ymin:ymax, xmin:xmax, 1] = transform_mat_func(g_weight_mat[element_id]) * g_scale
            else:
                result[ymin:ymax, xmin:xmax, 1] *= 0

            result[ymin:ymax, xmin:xmax, 2] *= 0
            if activities is not None:
                result[ymin:ymax, xmin:xmax, 2] += activities[element_id]

    return result



'''


from PymoNNto import *
import os
import sys

#def get_gene_string(gene):  # same implementation as interface_function get_gene_id
#    id = ''
#    for key, value in gene.items():
#        id += '#' + key + '$' + str(value)
#    return id + '#'

def execute_local_file(slave_file, genome):
    for arg in sys.argv:  # remove old
        if 'genome=' in arg:
            sys.argv.remove(arg)
    sys.argv.append('genome=' + get_gene_id(genome))  # add new

    py_file = open(slave_file, "r")
    execution_string = py_file.read()
    exec(execution_string)
    py_file.close()

class Evolution_Device():#one device per thread

    def __init__(self, device_string, parent):
        self.device_string = device_string
        self.parent = parent
        self.current_gene = None

    def score_processing(self, genome):
        score = self.get_score(genome)

        if score is not None:
            genome['score'] = score
            self.new_score_event(genome)
        else:
            self.error_event(genome, 'loaded score is None/ not able to load score')

    def initialize_device_group(self):# override / called before initialize but only once per thread group
        return

    def initialize(self):# override
        return

    def main_loop_update(self):# override
        return

    def start(self):# override
        return

    def stop(self):# override
        return

    def new_score_event(self, genome):
        if genome['gen'] == self.parent.Breed_And_Select.generation:  # part of current execution? still relevant?
            self.parent.new_score_event(genome)
        else:
            self.parent.error_event(genome, 'processed gene result from previous generation')

    def error_event(self, genome, message):
        if genome['gen'] == self.parent.Breed_And_Select.generation: #part of current execution? still relevant?
            self.parent.error_event(genome, message)
        else:
            self.parent.error_event(genome, message+' (from previous generation)')

    def get_score(self, genome):# override
        return


from PymoNNto.Exploration.Network_UI.TabBase import *

class multi_group_plot_tab(TabBase):

    def __init__(self, variables, title='Multi Group', timesteps=500):
        super().__init__(title)
        self.timesteps = timesteps

        self.original_variables=variables
        self.variables = [] #{var1:0, var2:0, var3:1, variable:group}
        self.curve_numbers = []
        self.plot_variables = []
        for i, var in enumerate(self.original_variables):#grouped togerther with var1|var2
            splitted=var.split('|')
            for v in splitted:
                self.variables.append(v)
            self.curve_numbers.append(len(splitted))
            self.plot_variables.append(splitted)


    def add_recorder_variables(self, neuron_group, Network_UI):
        for var in self.variables:
            Network_UI.add_recording_variable(neuron_group, 'np.mean(n.' + var + ')', timesteps=self.timesteps)
            Network_UI.add_recording_variable(neuron_group, 'n.'+var, timesteps=self.timesteps)



    def initialize(self, Network_UI):
        self.main_tab = Network_UI.Next_Tab(self.title)

        group_count = len(Network_UI.neuron_visible_groups)

        self.net_plot_dicts = []

        for plot_id, plot_variable_list in enumerate(self.plot_variables):

            curves_per_plot = len(plot_variable_list)

            stretch = 1
            if plot_id == 0:
                stretch = 2

            curves = Network_UI.Add_plot_curve(stretch=stretch,
                                               number_of_curves=group_count*curves_per_plot,
                                               return_list=True,
                                               x_label='t (iterations)',
                                               y_label='Network average ' + str(plot_variable_list)) #, lines=lines

            ci=0
            plot_curve_dict = {}
            for var in plot_variable_list:
                variable_curves = []
                for group_index in range(group_count):
                    variable_curves.append(curves[ci])
                    ci += 1
                plot_curve_dict[var] = variable_curves
            self.net_plot_dicts.append(plot_curve_dict)


        Network_UI.Next_H_Block()

        self.neuron_plot_dicts = []

        for plot_id, plot_variable_list in enumerate(self.plot_variables):

            curves_per_plot = len(plot_variable_list)

            stretch = 1
            if plot_id == 0:
                stretch = 2

            curves = Network_UI.Add_plot_curve(stretch=stretch,
                                               number_of_curves=1 + curves_per_plot,
                                               colors=[Network_UI.neuron_select_color],
                                               legend=False,
                                               x_label='t (iterations)',
                                               y_label='Neuron ' + str(plot_variable_list))

            ci=0
            plot_curve_dict = {}
            for var in plot_variable_list:
                plot_curve_dict[var] = curves[ci]
                ci += 1
            self.neuron_plot_dicts.append(plot_curve_dict)



        if Network_UI.group_display_count > 1:
            Network_UI.Next_H_Block()

            self.group_sliders = []
            for group_index in range(Network_UI.group_display_count):
                self.group_sliders.append(QSlider(1))  # QtCore.Horizontal
                self.group_sliders[-1].setMinimum(0)
                self.group_sliders[-1].setMaximum(100)
                self.group_sliders[-1].setSliderPosition(100)
                self.group_sliders[-1].mouseReleaseEvent = Network_UI.static_update_func
                self.group_sliders[-1].setToolTip('scale neuron-group plots up and down (only visualization)')

                Network_UI.Add_element(self.group_sliders[-1])


    def update(self, Network_UI):
        if self.main_tab.isVisible():

            lg=len(Network_UI.neuron_visible_groups)
            for curve_dict in self.net_plot_dicts:
                for variable in curve_dict:
                    for group_id, curve in enumerate(curve_dict[variable]):
                        group_tag=Network_UI.neuron_visible_groups[group_id]
                        group = Network_UI.network[group_tag, 0]

                        if hasattr(self, 'group_sliders'):
                            squeeze = self.group_sliders[group_id].sliderPosition() / 100
                        else:
                            squeeze = 1

                        try:
                            net_data = group['np.mean(n.' + variable + ')', 0, 'np'][-self.timesteps:]
                            iterations = group['n.iteration', 0, 'np'][-self.timesteps:]
                            curve.setData(iterations, net_data * squeeze, pen=group.color)
                        except:  # else:
                            curve.clear()

            group = Network_UI.selected_neuron_group()
            for curve_dict in self.neuron_plot_dicts:
                for variable in curve_dict:
                    curve = curve_dict[variable]

                    try:
                        neuron_data = group['n.' + variable, 0, 'np'][-self.timesteps:].astype(def_dtype)#for bool
                        iterations = group['n.iteration', 0, 'np'][-self.timesteps:]
                        if len(neuron_data.shape) > 1:
                            neuron_data = np.mean(neuron_data[:, Network_UI.selected_neuron_mask()], axis=1)
                        curve.setData(iterations, neuron_data)
                    except:
                        curve.clear()



from PymoNNto.NetworkCore.Behaviour import *

class Instant_Homeostasis(Behaviour):

    def get_measurement_param(self, n):
        if self.compiled_mp is None:
            self.compiled_mp = compile(self.measurement_param, '<string>', 'eval')

        result = eval(self.compiled_mp)

        if self.measurement_min is not None: result = np.clip(result, self.measurement_min, None)
        if self.measurement_max is not None: result = np.clip(result, None, self.measurement_max)
        return result

    def add_to_target_param(self, neurons, add):
        val = getattr(neurons, self.adjustment_param)+add
        if self.target_clip_min is not None or self.target_clip_max is not None:
            val = np.clip(val, self.target_clip_min, self.target_clip_max)
        setattr(neurons, self.adjustment_param, val)


    def get_target_adjustment(self, measure_or_average):

        greater = (measure_or_average > self.max_th) * (-self.dec)
        smaller = (measure_or_average < self.min_th) * self.inc

        if self.distance_sensitive:
            greater *= measure_or_average - self.max_th
            smaller *= self.min_th - measure_or_average

        return (greater+smaller)*self.adj_strength

    def set_threshold(self, th, gap_percent=None):
        self.th = th #just for readout and copying

        if th is not None:
            self.min_th = th
            self.max_th = th

            if gap_percent is not None:
                self.min_th -= self.min_th/100*gap_percent
                self.max_th += self.max_th/100*gap_percent


    def set_variables(self, neurons):

        self.compiled_mp=None

        self.min_th = self.get_init_attr('min_th', None, neurons)                           #minumum threshold for measurement param
        self.max_th = self.get_init_attr('max_th', None, neurons)                           #maximum threshold for measurement param

        self.set_threshold(                                                                 #optional
            self.get_init_attr('threshold', None, neurons),                                 #threshold for measurement param (min=max=th)
            self.get_init_attr('gap_percent', None, neurons)                                #min max gap is created via a percentage from th (additional param for th)
        )

        self.distance_sensitive = self.get_init_attr('distance_sensitive', True, neurons)   #stronger adjustment when value is further away from optimum

        self.inc = self.get_init_attr('inc', 1.0, neurons)                                  #increase factor
        self.dec = self.get_init_attr('dec', 1.0, neurons)                                  #decrease factor
        self.adj_strength = self.get_init_attr('adj_strength', 1.0, neurons)                #change factor

        self.adjustment_param = self.get_init_attr('adjustment_param', None, neurons)       #name of neurons target attribute

        self.measurement_param = self.get_init_attr('measurement_param', None, neurons)     #name of parameter to be measured

        self.measurement_min = self.get_init_attr('measurement_min', None, neurons)         #minumum value which can be measured (below=0)
        self.measurement_max = self.get_init_attr('measurement_max', None, neurons)         #maximum value which can be measured (above=max)

        self.target_clip_min = self.get_init_attr('target_clip_min', None, neurons)         #target clip min
        self.target_clip_max = self.get_init_attr('target_clip_max', None, neurons)         #target clip max


    def new_iteration(self, neurons):
        measure_or_average = self.get_measurement_param(neurons)
        self.add = self.get_target_adjustment(measure_or_average)
        self.add_to_target_param(neurons, self.add)


class Time_Integration_Homeostasis(Instant_Homeostasis):

    def get_measurement_param(self, n):

        value = super().get_measurement_param(n)
        self.average = (self.average * self.integration_length + value) / (self.integration_length+1)
        return self.average

    def set_variables(self, neurons):
        super().set_variables(neurons)

        self.integration_length = self.get_init_attr('integration_length', 1, neurons)      #factor that determines the inertia of the leaky integrator (0=instant) (a*i)/(1+i)

        self.average = self.get_init_attr('init_avg', (self.min_th+self.max_th)/2, neurons)                          #initialize average measurement for each neuron

#Time_Integration_Homeostasis().visualize_module(vmi=['measure'], vmo=['adjust'], vma=['measurement_param','adjustment_param','threshold','integration_length','...'])


from PymoNNto.NetworkCore.Behaviour import *
from PymoNNto.NetworkBehaviour.EulerEquationModules.Helper import *

class Variable(Behaviour):

    def set_variables(self, neurons):
        n = neurons

        eq_parts = eq_split(self.get_init_attr('eq', None))

        if eq_parts[1] == '=' and len(eq_parts) >= 3:
            self.var_name = eq_parts[0]
        else:
            print('invalid formula')

        self.add_tag('Variable '+self.var_name)

        eq_parts = remove_units(eq_parts, 2)

        self.var_init = ''.join(eq_parts[2:])

        setattr(n, self.var_name, neurons.get_neuron_vec() + eval(self.var_init))#symbols('n.'+self.var_name) #+

        setattr(n, self.var_name+'_new', neurons.get_neuron_vec() + eval(self.var_init))

    def new_iteration(self, n):
        setattr(n, self.var_name, getattr(n, self.var_name+'_new'))#apply the new value to variable


class Syn_Variable(Behaviour):

    def set_variables(self, synapse):
        s = synapse

        eq_parts = eq_split(self.get_init_attr('eq', None))

        if eq_parts[1] == '=' and len(eq_parts) >= 3:
            self.var_name = eq_parts[0]
        else:
            print('invalid formula')

        self.add_tag('Syn_Variable '+self.var_name)

        eq_parts = remove_units(eq_parts, 2)

        self.var_init = ''.join(eq_parts[2:])

        setattr(s, self.var_name, synapse.get_synapse_mat() + eval(self.var_init))#symbols('n.'+self.var_name) #+

        setattr(s, self.var_name+'_new', synapse.get_synapse_mat() + eval(self.var_init))

    def new_iteration(self, s):
        setattr(s, self.var_name, getattr(s, self.var_name+'_new'))#apply the new value to variable




import numpy as np
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt

#X = np.array([np.random.rand(1000) for i in range(1000)])

def get_PCA(act, num_sv=100):
    num_sv = min(num_sv, len(act))
    pca = PCA(n_components=num_sv)
    pca.fit(act)

    return pca

def get_activity_singular_values(act, num_sv=100):
    num_sv = min(num_sv, len(act))
    pca = PCA(n_components=num_sv)
    pca.fit(act)

    return pca.singular_values_

def get_activity_singular_values_and_components(act, num_sv=100):
    num_sv = min(num_sv, len(act))
    pca = PCA(n_components=num_sv)
    pca.fit(act)

    return pca.singular_values_, pca.components_, pca.explained_variance_, pca.explained_variance_ratio_


#pca.explained_variance_ !!!




from PymoNNto import *




def get_unique_non_partitioned_Groups(groups):
    temp = {}
    result = []
    for g in groups:
        key = ','.join(g.tags)
        if key not in temp:
            temp[key] = [g]
            result.append(g)
        else:
            temp[key].append(g)
    return result


#########################################
####Partitioned
#########################################

def get_partitioned_synapse_matrix(neurons, synapse_tag, synapse_var, return_first=True):
    results = {}
    shapes = {}
    for s in neurons.afferent_synapses[synapse_tag]:
        base_src = s.src.group_without_subGroup()
        base_dst = s.dst.group_without_subGroup()
        key = ','.join(s.tags)
        if not key in results:
            results[key] = np.zeros((base_dst.size, base_src.size))
            shapes[key] = (base_src.height, base_src.width)
        try:
            syn_mat = eval('s.' + synapse_var)
            if base_src == s.src and base_dst == s.dst:
                results[key] += syn_mat  # .copy()
            else:
                mat_mask = s.dst.mask[:, None] * s.src.mask[None, :]
                results[key][mat_mask] += np.array(syn_mat).flatten()  # np.array required if syn_mat is bool (enabled)
        except:
            print(synapse_var, "cannot be evaluated")

    if return_first:
        return list(results.values())[0]
    else:
        return results

def set_partitioned_synapse_matrix(neurons, synapse_tag, synapse_var, mat):#warning! synapse_tag has to be unique and only used by partition synapses
    for s in neurons.afferent_synapses[synapse_tag]:
        try:
            mask = s.dst.mask[:, None] * s.src.mask[None, :]
            setattr(s, synapse_var, mat[mask].reshape(s.get_synapse_mat_dim()))
        except:
            print(synapse_var, "cannot be set")


def SingleNeuron_attached_SubSGs(neuron_group, synapse_tag, neuron_id):
    result = []
    for s in neuron_group.afferent_synapses[synapse_tag]:
        if neuron_id in s.dst.id:
            result.append(s)
    return result

def get_single_neuron_combined_partition_matrix(neurons, synapse_tag, synapse_var, neuron_id, return_first=True):
    results = []
    for s in SingleNeuron_attached_SubSGs(neurons, synapse_tag, neuron_id):
        base_src = s.src.group_without_subGroup()
        mat = np.zeros(base_src.height*base_src.width)
        indx = np.where(s.dst.id == neuron_id)[0]
        mat[s.src.mask] = eval('s.' + synapse_var)[indx].flatten()
        results.append(mat.reshape((base_src.height, base_src.width)))

    if return_first:
        return results[0]
    else:
        return results

#def set_partitioned_single_neuron_weights(neurons, synapse_tag, synapse_var, neuron_id, vec):
#    for s in SingleNeuron_attached_SubSGs(neurons, synapse_tag, id):
#        base_src = s.src.group_without_subGroup()
#        mat = np.zeros(base_src.height*base_src.width)
#        mat[s.src.mask] = eval('s.' + synapse_var).flatten()
#        results.append(mat.reshape((base_src.height, base_src.width)))

#        try:
#            mask = s.dst.mask[:, None] * s.src.mask[None, :]
#            setattr(s, synapse_var, mat[mask].reshape(s.get_synapse_mat_dim()))
#        except:
#            print(synapse_var, "cannot be set")


#def get_partitioned_matrix(neuron, synapse_tag, synapse_var):
#    return list(get_combined_syn_mats(neuron.afferent_synapses[synapse_tag], None, synapse_var).values())[0]

#def set_partitioned_matrix(neuron, synapse_tag, synapse_var, mat):



#########################################
####Old
#########################################


def get_combined_syn_mats(synapses, neuron_id=None, attr='W'):
    results = {}
    shapes = {}
    for s in synapses:
        base_src = s.src.group_without_subGroup()
        base_dst = s.dst.group_without_subGroup()
        key = ','.join(s.tags)
        if not key in results:
            results[key] = np.zeros((base_dst.size, base_src.size))
            shapes[key] = (base_src.height, base_src.width)
        try:
            syn_mat = eval('s.' + attr)
            if base_src == s.src and base_dst == s.dst:
                results[key] += syn_mat  # .copy()
            else:
                mat_mask = s.dst.mask[:, None] * s.src.mask[None, :]
                results[key][mat_mask] += np.array(syn_mat).flatten()  # np.array required if syn_mat is bool (enabled)
        except:
            print(attr, "cannot be evaluated")

    if neuron_id is not None:
        for key in results:
            results[key] = results[key][neuron_id].reshape(shapes[key])

    return results




from PymoNNto.Exploration.UI_Base import *
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import copy
import sys
import traceback
from PymoNNto.NetworkCore.Neuron_Group import *
from PymoNNto.NetworkCore.Synapse_Group import *

from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject

import inspect

class Module_draw_item(pg.GraphicsObject):

    def __init__(self, module_name, inputs, outputs, attributes, module_type, init_vars=None, x=0, y=0, onlyff=False):
        #super().__init__(None)
        GraphicsObject.__init__(self)

        self.selected = True

        self.picture = QtGui.QPicture()

        self.draw_forward_only = onlyff
        self.draw_in_out_blocks = True
        self.draw_connection_lines = True
        self.draw_attributes = True

        self.conn_points = []

        self.size = 1000

        self.x = x
        self.y = y

        self.border = 0.11
        self.nob_size = 0.055

        self.module_name = module_name

        if init_vars is not None:
            self.init_vars = init_vars
        else:
            self.init_vars = [var.replace('#init', '') for var in sorted(outputs) if '#init' in var]

        self.outputs = [var.replace('#init', '') for var in sorted(outputs)]

        self.inputs = sorted(inputs)

        self.get_arrow_positions()

        self.attributes = sorted(attributes)
        self.module_type = module_type

        self.red = (237, 85, 59, 255)
        self.green = (60, 174, 163, 255)
        self.blue = (32, 99, 155, 255)
        self.yellow = (246, 213, 92, 255)

    def get_position(self, column_element_count, element_id, x_pos): #left, right, center

        if x_pos == 'left':
            x = self.x+self.border * self.size
            d=4.5

        if x_pos == 'right':
            x = self.x+ self.size - self.border * self.size
            d=4.5

        if x_pos == 'center':
            x = self.x
            d=4.5

        shift = 0
        h = self.size - self.border * self.size * d

        if column_element_count > 1:
            shift = (h / (column_element_count-1) * element_id)
        else:
            shift = h / 2.0

        y = self.y -(self.border * self.size * 1.5 + shift)
        #y = -self.border * self.size * 2 + ((1 - self.border * self.size * 4) / column_element_count * element_id)

        return x, y

    def set_font_smaller_than(self, painter, text, max_width, text_size):
        tw=max_width

        while tw >= max_width:
            qf = QFont("Arial")
            qf.setPointSizeF(text_size)

            text_size -= 0.1

            fm = QFontMetrics(qf)
            # qf.setStretch(2)
            painter.setFont(qf)
            tw = fm.width(text)

        return fm

    def draw_arrow(self, painter, pos, text, color):
        x=pos.x()
        y=pos.y()


        painter.setPen(pg.mkPen(color=color))
        painter.setBrush(pg.mkBrush(color=color))

        # painter.drawRect(QtCore.QRectF(x, y, self.nob_size * self.size, self.nob_size * self.size))

        points = QPolygonF([
            QtCore.QPointF(x - self.nob_size * self.size * 3, y - self.nob_size * self.size * 0.7),
            QtCore.QPointF(x + self.nob_size * self.size * 2, y - self.nob_size * self.size * 0.7),

            QtCore.QPointF(x + self.nob_size * self.size * 2 + self.nob_size * self.size * 0.7, y),  # arrow

            QtCore.QPointF(x + self.nob_size * self.size * 2, y + self.nob_size * self.size * 0.7),
            QtCore.QPointF(x - self.nob_size * self.size * 3, y + self.nob_size * self.size * 0.7),
            # QtCore.QPointF(x+100, y+100)
        ])

        painter.drawPolygon(points)

        painter.setPen(pg.mkPen(color=(0, 0, 0, 255)))
        painter.setBrush(pg.mkBrush(color=(0, 0, 0, 255)))

        fm = self.set_font_smaller_than(painter, text, self.nob_size * self.size * 5, self.size / 100.0 * 4.0)

        #text_size=4.0
        #tw = self.nob_size * self.size * 5

        #while tw >= self.nob_size * self.size * 5:
        #    qf = QFont("Arial")
        #    qf.setPointSizeF(self.size / 100.0 * text_size)

        #    text_size -= 0.1

        #    fm = QFontMetrics(qf)
        #    # qf.setStretch(2)
        #    painter.setFont(qf)
        #    tw = fm.width(text)


        painter.drawText(QtCore.QPointF(x - fm.width(text) / 2.0, y + fm.height() / 4), text)  # + self.nob_size * self.size / 2  float(fm.height())/20.0

        # axis.add_patch(Rectangle((x, y), nob_size, nob_size, alpha=1, color=c))
        # axis.add_patch(Circle((x + nob_size, y + nob_size / 2), 1.0 * nob_size, alpha=1, color=c))
        # axis.text(x + nob_size, y + nob_size / 2, s, fontsize=12, ha='right', va='center')

    def get_arrow_positions(self):
        self.in_pos=[]

        for i, s in enumerate(self.inputs):
            x, y = self.get_position(len(self.inputs), i, 'left')
            self.in_pos.append(QtCore.QPointF(x,y))

        self.out_pos = []
        for i, s in enumerate(self.outputs):
            x, y = self.get_position(len(self.outputs), i, 'right')
            self.out_pos.append(QtCore.QPointF(x, y))

    def add_in_out_blocks(self, painter, input_labels, output_labels):
        for s, pos in zip(self.inputs, self.in_pos):
            if s in output_labels:
                color = self.yellow
            else:
                color = self.red
            self.draw_arrow(painter, pos, s, color)

        for s, pos in zip(self.outputs,self.out_pos):
            if s in input_labels and not s in self.init_vars:
                color = self.yellow
            else:
                color = self.green
            self.draw_arrow(painter, pos, s, color)



    def draw_content(self, painter, labels):

        for i, s in enumerate(labels):

            x, y = self.get_position(len(labels), i, 'center')
            #y = -self.border * self.size * 2 + ((1 - self.border * self.size * 4) / len(labels) * i)

            color = (255, 255, 255, 255)
            painter.setPen(pg.mkPen(color=color))
            painter.setBrush(pg.mkBrush(color=color))

            qf = QFont("Arial")
            qf.setPointSizeF(self.size/100.0*4.0)

            #if ":" not in s and s!='':
            #    qf.setUnderline(True)
            #else:
            #    s='-'+s

            fm = QFontMetrics(qf)
            # qf.setStretch(2)
            painter.setFont(qf)

            painter.drawText(QtCore.QPointF(x + self.size/2-fm.width(s)/2, y+fm.height()/4), s) # + self.nob_size * self.size / 2

    def draw_connection(self, painter, from_pos, to_pos, distance, h, color):
        if distance > 0 or not self.draw_forward_only:
            from_right = QtCore.QPointF(from_pos.x() + self.nob_size * self.size * 1.0, from_pos.y())
            to_left = QtCore.QPointF(to_pos.x() - self.nob_size * self.size * 1.0, to_pos.y())
            center = QtCore.QPointF((from_pos.x() + to_pos.x()) / 2.0, (from_pos.y() + to_pos.y()) / 2.0)

            if distance != 1:

                if distance == 0:#for auto connections
                    distance=2

                if h > 0:
                    add = 200 * abs(distance)  # -200*h
                else:
                    add = -300 - 200 * abs(distance)  # -200*h

                center.setY(center.y() + add)
                from_right_2 = QtCore.QPointF(from_right.x(), (center.y() + from_right.y()) / 2.0)
                to_left_2 = QtCore.QPointF(to_left.x(), (center.y() + to_left.y()) / 2.0)
                center_left = QtCore.QPointF(from_pos.x() + self.nob_size * self.size * 2.0, center.y())
                center_right = QtCore.QPointF(to_pos.x() - self.nob_size * self.size * 2.0, center.y())

                #painter.drawLine(from_pos, from_right)
                #painter.drawLine(from_right, from_right_2)
                #painter.drawLine(from_right_2, center_left)
                #painter.drawLine(center_left, center)
                #painter.drawLine(center, center_right)
                #painter.drawLine(center_right, to_left_2)
                #painter.drawLine(to_left_2, to_left)
                #painter.drawLine(to_left, to_pos)

                pp = QPainterPath(from_pos)
                pp.quadTo(from_right, from_right_2)
                pp.quadTo(center_left, center)
                pp.quadTo(center_right, to_left_2)
                pp.quadTo(to_left, to_pos)
            else:
                pp = QPainterPath(from_pos)
                pp.quadTo(from_right, center)
                pp.quadTo(to_left, to_pos)

            if self.selected:
                color = (color[0], color[1], color[2], 255)
            else:
                color = (color[0], color[1], color[2], 100)

            painter.strokePath(pp, pg.mkPen(color=color, width=5))

    def draw_connections(self, painter, other_modules):
        points = []
        if other_modules is not None:
            self.self_index = other_modules.index(self)

            for in_indx, inp in enumerate(self.inputs):

                found = None

                #for f in range(self.self_index):
                for j in range(len(other_modules)-1):
                    f = self.self_index+1+j
                    if f>=len(other_modules)-1:
                        f=f-len(other_modules)

                    if inp in other_modules[f].outputs:
                        found = other_modules[f]


                if found is None:
                    if inp in self.outputs:
                        found = self


                if found is not None:

                    if inp in found.outputs and hasattr(found, 'out_pos'):
                        out_indx = found.outputs.index(inp)

                        from_pos = copy.copy(found.out_pos[out_indx])
                        from_pos.setX(from_pos.x()+self.nob_size*self.size*2.7)
                        to_pos = copy.copy(self.in_pos[in_indx])
                        to_pos.setX(to_pos.x()-self.nob_size*self.size*3)

                        points.append(from_pos)
                        distance = self.self_index - other_modules.index(found)

                        out_h = (len(found.outputs) + 1) / 2 - (found.outputs.index(inp) + 1)
                        inp_h = (len(self.inputs) + 1) / 2 - (self.inputs.index(inp) + 1)
                        h = out_h + inp_h

                        color = self.yellow
                        if inp in found.init_vars:
                            color = self.green
                            if found == self:
                                color = (0, 0, 0, 50)

                        self.draw_connection(painter, from_pos, to_pos, distance, h, color)

        return points


    def update_pic(self, other_modules=None):
        self.picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.picture)
        painter.scale(1, -1)

        if self.draw_connection_lines:
            self.conn_points = self.draw_connections(painter, other_modules)

        color = self.blue
        #if self.selected:
        #    painter.setPen(pg.mkPen(color=(0,0,0,255), width=10))
        #else:

        if self.selected:
            color = (color[0], color[1], color[2], 255)
        else:
            color = (color[0], color[1], color[2], 100)

        painter.setPen(pg.mkPen(color=(0, 0, 0, 0)))
        painter.setBrush(pg.mkBrush(color=color))

        #currentAxis.plot([0, 0, 1.0 + border], [0, 1, 1], color='white')

        painter.drawRect(QtCore.QRectF(self.x+self.border*self.size, self.y-self.size+self.border*self.size, self.size - self.border * self.size * 2, self.size - self.border * self.size * 2))

        color = (255, 255, 255, 255)
        painter.setPen(pg.mkPen(color=color))
        painter.setBrush(pg.mkBrush(color=color))



        text = self.module_name.replace('_', ' ')
        fm = self.set_font_smaller_than(painter, text, self.size - self.border * self.size * 3, text_size=self.size/100.0*7.0)

        #qf = QFont("Arial")
        #qf.setPointSizeF(self.size/100.0*7.0)
        #qf.setUnderline(True)
        #qf.setBold(True)
        #fm = QFontMetrics(qf)
        # qf.setStretch(2)
        #painter.setFont(qf)

        if self.draw_attributes:
            y = self.y-int(self.size - self.border * self.size * 1.5-fm.height()/2)
        else:
            y = self.y-int(self.size/2-fm.height()/2)

        painter.drawText(int(self.x+0.5*self.size-fm.width(text)/2), y, text)

        qf = QFont("Arial")
        #qf.setLetterSpacing(QFont.PercentageSpacing, 95)
        qf.setPointSizeF(self.size/100.0*1.5)
        fm = QFontMetrics(qf)
        # qf.setStretch(2)
        painter.setFont(qf)
        text = self.module_type
        painter.drawText(int(self.x+0.5 * self.size - fm.width(text) / 2), int(self.y-self.border * self.size*1.2 + fm.height() / 2), text)

        if self.draw_in_out_blocks:
            self.add_in_out_blocks(painter, self.inputs, self.outputs)

        if len(self.attributes) > 0 and self.draw_attributes:
            self.draw_content(painter, ['', '']+self.attributes+['', 'Attributes:'])#Initialization a



        painter.end()
        self.prepareGeometryChange()
        self.informViewBoundsChanged()


    def paint(self, painter, option, widget=None):
        painter.drawPicture(0, 0, self.picture)

    def boundingRect(self):
        br = self.picture.boundingRect()
        br.setLeft(self.x)  # min(br.top()-10, br.left()-10)
        br.setTop(-self.y+self.border*self.size)#max(br.top()-10, br.left()-10)
        br.setRight(self.x+self.size)#,max(br.right()+10, br.bottom()+10)
        br.setBottom(-self.y+self.size-self.border*self.size)#,max(br.right()+10, br.bottom()+10)

        #for p in self.conn_points:
        #    if p.x() < br.left(): br.setLeft(p.x())
        #    if p.y() < br.top(): br.setTop(p.y())
        #    if p.x() > br.right(): br.setRight(p.x())
        #    if p.y() > br.bottom(): br.setBottom(p.y())

        #print(br.left(), br.top(), br.right(), br.bottom())
        return QtCore.QRectF(br)

def analyze_module_and_get_info(module):
    from PymoNNto.NetworkCore.Neuron_Group import NeuronGroup
    from PymoNNto.NetworkCore.Synapse_Group import SynapseGroup

    module_copy = copy.deepcopy(module)

    ng = NeuronGroup_read_write_event(1, {}, None)
    sg = SynapseGroup_read_write_event(ng, ng, None)

    main_arg = ng
    module_type = 'PymoNNto Neuron-Group Module'
    synapse_beh_arg_names = ['s', 'S', 'syn', 'Syn', 'SYN', 'syns', 'Syns', 'SYNS', 'synapse', 'Synapse', 'SYNAPSE',
                             'synapses', 'Synapses', 'SYNAPSES']
    func1_arg_names = inspect.getfullargspec(module_copy.set_variables).args
    func2_arg_names = inspect.getfullargspec(module_copy.new_iteration).args
    for s in synapse_beh_arg_names:
        if s in func1_arg_names or s in func2_arg_names:
            main_arg = sg
            module_type = 'PymoNNto Synapse-Group Module'

    if module_copy.set_variables_on_init:
        module_type += ' (init set var)'

    # ng = NeuronGroup(1, {}, None)
    # sg = SynapseGroup(ng, ng, None)

    ng.afferent_synapses['All'] = [sg]
    ng.efferent_synapses['All'] = [sg]

    # print(module_copy, ng.__dict__)

    def reset_vars():
        ng.variable_reads = []
        ng.variable_writes = []
        sg.variable_reads = []
        sg.variable_writes = []

    reset_vars()

    # delattr(sg, 'enabled')

    def read_ng(obj, attr):
        obj.variable_reads.append('n.' + attr)

    def write_ng(obj, attr):
        obj.variable_writes.append('n.' + attr)

    def read_sg(obj, attr):
        obj.variable_reads.append('s.' + attr)

    def write_sg(obj, attr):
        obj.variable_writes.append('s.' + attr)

    # set_read_event_function(ng, read_ng)
    # set_write_event_function(ng, write_ng)

    # set_read_event_function(sg, read_sg)
    # set_write_event_function(sg, write_sg)

    ng.set_read_event_function(read_ng)
    ng.set_write_event_function(write_ng)

    sg.set_read_event_function(read_sg)
    sg.set_write_event_function(write_sg)

    created_vars_set = analyze_function(module_copy, 'set_variables', ng, sg, main_arg)
    write_vars_set = ng.variable_writes + sg.variable_writes  # [var for var in ng.variable_writes + sg.variable_writes if var not in created_vars_set]
    read_vars_set = ng.variable_reads + sg.variable_reads  # [var for var in ng.variable_reads + sg.variable_reads if var not in created_vars_set]

    # print(module_copy, created_vars_set)

    reset_vars()

    created_vars_it = analyze_function(module_copy, 'new_iteration', ng, sg, main_arg)
    write_vars_it = ng.variable_writes + sg.variable_writes  # [var for var in ng.variable_writes + sg.variable_writes if var not in created_vars_it]
    read_vars_it = ng.variable_reads + sg.variable_reads  # [var for var in ng.variable_reads + sg.variable_reads if var not in created_vars_it]

    inputs = list(set(read_vars_it + write_vars_it + created_vars_it))  # read_vars_set
    outputs_temp = list(set(
        write_vars_set + write_vars_it + created_vars_set + created_vars_it))  # created_vars_it #[var+'#init' for var in created_vars_set]

    outputs = []
    for o in outputs_temp:
        if o in created_vars_set:
            outputs.append(o + '#init')
            # print(module_copy, o)
        else:
            outputs.append(o)
        # outputs = [var+'#init' for var in outputs if var in created_vars_set]

    # inputs = list(set(ng.variable_reads + sg.variable_reads + created_vars_it))
    # outputs = list(set(ng.variable_writes + sg.variable_writes+created_vars_set + created_vars_it))

    module_name = module_copy.__class__.__name__

    attributes = module_copy.used_attr_keys.copy()  # ['transmitter', 'test', 'blub', 'param', 'dfsfa']
    if 'tag' in attributes:
        attributes.remove('tag')
    if 'behaviour_enabled' in attributes:
        attributes.remove('behaviour_enabled')

    attributes = list(set(attributes))

    if hasattr(module_copy, 'visualization_module_inputs') and module_copy.visualization_module_inputs is not None:
        inputs = module_copy.visualization_module_inputs

    if hasattr(module_copy, 'visualization_module_outputs') and module_copy.visualization_module_outputs is not None:
        outputs = module_copy.visualization_module_outputs

    if hasattr(module_copy, 'visualization_module_attributes') and module_copy.visualization_module_attributes is not None:
        attributes = module_copy.visualization_module_attributes

    # filter
    filter = ['n.iteration', 's.iteration']
    inputs = [var for var in inputs if not var in filter]
    outputs = [var for var in outputs if not var in filter]
    attributes = [var for var in attributes if not var in filter]

    return module_name, inputs, outputs, attributes, module_type


class module_drawer(UI_Base):

    def __init__(self):
        super().__init__(None, label='Drawer', create_sidebar=False)

        pg.setConfigOptions(antialias=True)

        self.main_window.setGeometry(10, 40, 714, 608)

        self.di_modules = []

    def add_flow_chart(self, ff_only=False):
        self.flow_tab = self.Next_Tab('Flow_Chart', stretch=0.0)
        self.flow_tab.plot = self.Add_plot()
        self.flow_tab.plot.hideAxis('bottom')
        self.flow_tab.plot.hideAxis('left')
        self.flow_tab.modules = []

        self.x = 0
        self.y = 0

        def on_plot_click_event(event):
            for mod in self.flow_tab.modules:
                mod.selected = True

            for mod in self.flow_tab.modules:
                mod.update_pic(self.flow_tab.modules)

            self.flow_tab.plot.update()

        self.flow_tab.plot.mouseClickEvent = on_plot_click_event

        for mod in self.di_modules:
            mdi = Module_draw_item(mod.module_name, mod.inputs, mod.outputs, mod.attributes, mod.module_type, mod.init_vars, self.x, self.y, ff_only)

            self.flow_tab.modules.append(mdi)
            self.flow_tab.plot.addItem(mdi)

            def on_mod_click_event(event):
                for mod in self.flow_tab.modules:
                    mod.selected = False

                event.currentItem.selected = True

                for mod in self.flow_tab.modules:
                    mod.update_pic(self.flow_tab.modules)
                self.flow_tab.plot.update()

            mdi.mouseClickEvent = on_mod_click_event

            self.x += 1200
            # self.y += 1200

        for mod in self.flow_tab.modules:
            mod.update_pic(self.flow_tab.modules)
        self.flow_tab.plot.update()



    def add_module_tab(self, module_name, inputs, outputs, attributes, module_type):

        self.main_tab = self.Next_Tab(module_name, stretch=0.0)
        self.main_tab.plot = self.Add_plot()

        mdi = Module_draw_item(module_name, inputs, outputs, attributes, module_type)
        self.main_tab.draw_item = mdi
        self.di_modules.append(mdi)
        self.main_tab.plot.addItem(mdi)

        #def mce(event):
        #    print('dfsgsdfgdf')
        #mdi.mouseClickEvent = mce

        #self.main_tab.plot.setXRange(-5, 5)
        #self.main_tab.plot.sigXRangeChanged.connect(lambda _, t: self.main_tab.plot.setYRange(*t))

        self.main_tab.draw_item.update_pic(None)
        self.main_tab.plot.hideAxis('bottom')
        self.main_tab.plot.hideAxis('left')

        self.main_tab.plot.update()


    def add_module(self, module):
        module_name, inputs, outputs, attributes, module_type = analyze_module_and_get_info(module)
        self.add_module_tab(module_name, inputs, outputs, attributes, module_type)


def set_read_event_function(object, ref):
    object.current_dict = copy.copy(dir(object))
    object.ref = ref

    def __getattr__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        self.ref(self, attr_name)
        return super().__getattribute__(attr_name)#self.get_neuron_vec('uniform')

    def __getattribute__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        self.ref(self, attr_name)
        return super().__getattribute__(attr_name)#self.get_neuron_vec('uniform')

    object.__getattr__ = __getattr__
    object.__getattribute__ = __getattribute__


def set_write_event_function(object, wef):
    object.current_dict = copy.copy(dir(object))
    object.wef = wef

    def __setattr__(self, attr_name, val):
        if self.wef is not None:
            self.wef(self, attr_name)
        super().__setattr__(attr_name, val)

    object.__setattr__ = __setattr__


class NeuronGroup_read_write_event(NeuronGroup):

    ref = None
    wef = None
    current_dict = None

    def set_read_event_function(self, ref):
        self.current_dict = copy.copy(dir(self))
        self.ref = ref

    def set_write_event_function(self, wef):
        self.current_dict = copy.copy(dir(self))
        self.wef = wef

    def __getattr__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        self.ref(self, attr_name)
        return super().__getattribute__(attr_name)#self.get_neuron_vec('uniform')

    def __getattribute__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        self.ref(self, attr_name)
        return super().__getattribute__(attr_name)#self.get_neuron_vec('uniform')

    def __setattr__(self, attr_name, val):
        if self.wef is not None:
            self.wef(self, attr_name)
        super().__setattr__(attr_name, val)

class SynapseGroup_read_write_event(SynapseGroup):

    ref = None
    wef = None
    current_dict = None

    def set_read_event_function(self, ref):
        self.current_dict = copy.copy(dir(self))
        self.ref = ref

    def set_write_event_function(self, wef):
        self.current_dict = copy.copy(dir(self))
        self.wef = wef

    def __getattr__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        self.ref(self, attr_name)
        return super().__getattribute__(attr_name)#self.get_synapse_mat('uniform')

    def __getattribute__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        self.ref(self, attr_name)
        return super().__getattribute__(attr_name)#self.get_synapse_mat('uniform')

    def __setattr__(self, attr_name, val):
        if self.wef is not None:
            self.wef(self, attr_name)
        super().__setattr__(attr_name, val)

def analyze_function(object, function_name, ng, sg, arg):
    np.seterr(all='raise')

    ng_dict_backup = copy.copy(ng.__dict__)
    sg_dict_backup = copy.copy(sg.__dict__)

    finished = False
    used_variable_keys_ng = []
    used_variable_keys_sg = []

    error_txt = 'failed to analyse module automatically.'+ str(object) +' (You can set the modules attributes, input and output manually by adding the variables visualization_module_attributes, visualization_module_inputs and visualization_module_outputs (string lists) to the module manually)'

    break_counter=0
    while not finished:
        try:
            getattr(object, function_name)(arg)

            finished = True
        except Warning:
            pass
        except Exception as e:
            break_counter+=1
            if break_counter>30:
                #traceback.print_tb(e.__traceback__)
                print(error_txt)
                finished = True

            error_type = sys.exc_info()[0]
            msg = str(sys.exc_info()[1])
            #print(error_type, msg)
            if str(error_type.__name__) == 'AttributeError':
                sp = msg.split("'")
                if len(sp) > 3:
                    key = sp[3]

                    if 'SynapseGroup_read_write_event' in msg:
                        used_variable_keys_sg.append('s.'+key)
                        var = sg.get_synapse_mat()
                        setattr(sg, key, var)

                    elif 'NeuronGroup_read_write_event' in msg:
                        used_variable_keys_ng.append('n.'+key)
                        var = ng.get_neuron_vec()
                        setattr(ng, key, var)

            elif str(error_type.__name__) == 'KeyError':
                tag=msg.replace("'", '')
                ng.afferent_synapses[tag] = [sg]
                ng.efferent_synapses[tag] = [sg]

            elif 'invalid value encountered' in msg:
                pass
            else:
                #print(error_type, msg)
                #traceback.print_tb(e.__traceback__)
                print(error_txt)
                finished = True

    created_variables_ng = ['n.'+var for var in ng.__dict__ if var not in ng_dict_backup]
    created_variables_sg = ['s.'+var for var in sg.__dict__ if var not in sg_dict_backup]

    return created_variables_ng+created_variables_sg #used_variable_keys_ng+used_variable_keys_sg+








from PymoNNto import *

def behaviour_test_environment(behaviour, size=1, initialize=True, iteration=0):
    net = Network(tag='Network')

    if type(behaviour) != dict:
        behaviour={1:behaviour}

    ng = NeuronGroup(tag='Neuron', net=net, behaviour=behaviour, size=size)

    if initialize:
        net.initialize()

    for i in range(iteration):
        net.simulate_iteration()

    return net, ng, behaviour


from PymoNNto.NetworkCore.Neuron_Group import *


class NeuronGroup_read_write_event(NeuronGroup):

    ref = None
    wef = None
    current_dict = None

    def set_read_event_function(self, ref):
        self.current_dict = copy.copy(dir(self))
        self.ref = ref

    def set_write_event_function(self, wef):
        self.current_dict = copy.copy(dir(self))
        self.wef = wef

    def __getattr__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattr__(attr_name)
        print(attr_name+'-')
        if hasattr(self,attr_name):
            self.ref(self, attr_name+'-')
        return super().__getattr__(attr_name)#self.get_neuron_vec('uniform')

    def __getattribute__(self, attr_name):
        if attr_name in ['wef', 'ref', 'current_dict'] or self.current_dict is None or attr_name in self.current_dict:
            return super().__getattribute__(attr_name)
        #print(attr_name+'+')
        #if hasattr(self, attr_name):
        self.ref(self, attr_name+'+')
        return super().__getattribute__(attr_name)#self.get_neuron_vec('uniform')

    def __setattr__(self, attr_name, val):
        if self.wef is not None:
            self.wef(self, attr_name)
        super().__setattr__(attr_name, val)

def something_get(obj, attr):
    print('get', obj, attr)
    return

def something_set(obj, attr):
    print('set', obj, attr)
    return

NG_rwe = NeuronGroup_read_write_event(1, {}, None)

#print(NG_rwe.__doc__)

NG_rwe.set_read_event_function(something_get)
NG_rwe.set_write_event_function(something_set)

NG_rwe.afferent_synapses[1] = 2

print(hasattr(NG_rwe, 'x'))

#NG_rwe.x = np.zeros(10)
#y = NG_rwe.x + 10

#def dsfagsd(blub):
#    n=blub
#    return copy.copy(eval('np.mean(n.x)'))

#print(dsfagsd(NG_rwe))




from PymoNNto.Exploration.Network_UI.TabBase import *

class character_activation_tab(TabBase):

    def __init__(self, title='char'):
        super().__init__(title)
        self.title = title

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None:
            self.reconstruction_tab = Network_UI.Next_Tab(self.title)


            self.source=Network_UI.network['grammar_act', 0]

            _, self.plt = Network_UI.Add_plot_curve(title='Character activations', return_plot=True, number_of_curves=0, x_label='t (iterations)', y_label='Network average ')

            self.max_char_scatter = pg.ScatterPlotItem(pen=(100,100,100), brush=(100,100,100))
            self.max_char_scatter.activity_data = []

            self.char_scatter_plots=[]
            for i in range(len(self.source.alphabet)):
                symbol = QPainterPath()
                # symbol.addEllipse(QtCore.QRectF(-0.5, -0.5, 1, 1))

                font = QFont()
                #font.setStyleHint(QFont.Helvetica)
                font.setPointSize(1)
                char=self.source.index_to_char(i)
                if char==' ':
                    char='_'
                symbol.addText(-0.5, 0.2, font, char)

                color=set(np.random.rand(3)*255)
                self.char_scatter = pg.ScatterPlotItem(pen=color, brush=color, symbol=symbol)
                self.plt.addItem(self.char_scatter)
                self.char_scatter_plots.append(self.char_scatter)
                self.char_scatter.activity_data = []

            #self.data = np.zeros((len(source.alphabet),1))

            #self.timesteps=100


    def update(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None and self.reconstruction_tab.isVisible():
            group = Network_UI.selected_neuron_group()

            if hasattr(group, 'Input_Weights') and hasattr(group, 'output'):
                activation = group.Input_Weights.transpose().dot(group.output)

                self.max_char_scatter.activity_data.append(np.max(activation))
                if len(self.max_char_scatter.activity_data) > 50:
                    self.max_char_scatter.activity_data = self.max_char_scatter.activity_data[1:]

                iterations = group['n.iteration', 0, 'np'][-len(self.max_char_scatter.activity_data):]
                self.max_char_scatter.setData(iterations, self.max_char_scatter.activity_data, size=50)

                for i, char_scatter in enumerate(self.char_scatter_plots):
                    char_scatter.activity_data.append(activation[i])
                    if len(char_scatter.activity_data) > 50:
                        char_scatter.activity_data = char_scatter.activity_data[1:]

                    char_scatter.setData(iterations, char_scatter.activity_data, size=50)#list(range(len(char_scatter.activity_data)))



                #if len(neuron_data.shape) > 1:
                #    neuron_data = neuron_data[:, Network_UI.selected_neuron_id()]
                #self.neuron_var_curves[var].setData(iterations, neuron_data)

                #self.char_Act_plots.setData(self.data)


#import matplotlib
#matplotlib.use('Qt5Agg')

from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import numpy as np
#from Exploration.Visualization.Visualization_Helper import *

import sys
import PymoNNto

sys._excepthook = sys.excepthook
def exception_hook(exctype, value, traceback):
    print(exctype, value, traceback)
#    sys._excepthook(exctype, value, traceback)
    sys.exit(1)
sys.excepthook = exception_hook



class UI_Base(QApplication):

    def __init__(self, network, label="Network_Test", create_sidebar=True):
        super().__init__(sys.argv)

        path = PymoNNto.__file__.replace('__init__.py', '')+ 'icon3232.png'

        if 'win' in sys.platform and sys.platform != 'darwin':
            import ctypes
            myappid = 'mv.Pymonnto.ui.1'  # arbitrary string mycompany.myproduct.subproduct.version
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

        app_icon = QIcon()
        app_icon.addFile(path, QtCore.QSize(32, 32))
        self.setWindowIcon(app_icon)

        #app_icon.addFile(path + 'icon.png')
        #app_icon.addFile(path+'icon1616.png', QtCore.QSize(16, 16))
        #app_icon.addFile(path + 'icon2424.png', QtCore.QSize(24, 24))

        #app_icon.addFile(path + 'icon4848.png', QtCore.QSize(48, 48))
        #app_icon.addFile(path+'icon6464.png', QtCore.QSize(64, 64))
        #app_icon.addFile(path+'icon128128.png', QtCore.QSize(128, 128))
        #app_icon.addFile(path+'icon256256.png', QtCore.QSize(256, 256))


        #tray=QSystemTrayIcon(app_icon, parent=self)
        #tray.show()


        self.reduced_layout = False

        self.network = network

        self.main_window = QWidget()
        self.main_window.keyPressEvent = self.keyPressEvent
        self.main_window.keyReleaseEvent = self.keyReleaseEvent
        self.control_key_down=False

        self.init_QT_Window(label, create_sidebar)
        quit = QAction("Quit", self.main_window)
        quit.triggered.connect(self.main_window.closeEvent)
        self.closeeventtriggered = False

    def show(self):
        self.main_window.show()
        sys.exit(self.exec_())

    def Add_element(self, elem, sidebar=False, stretch=1):
        if sidebar:
            self.Add_Sidebar_Element(elem)
        else:
            self.current_H_block.addWidget(elem, stretch=stretch)
        return elem

    def Add_plot(self, title=None, sidebar=False, stretch=1, axisItems=None, x_label=None, y_label=None, always_show_axis_labels=False, tooltip_message=None):
        canvas = pg.GraphicsLayoutWidget()
        canvas.setBackground((255, 255, 255))

        if tooltip_message is not None:
            canvas.ci.setToolTip(tooltip_message)

        self.Add_element(canvas, sidebar, stretch)

        plt = canvas.addPlot(row=0, col=0, axisItems=axisItems)

        if always_show_axis_labels or not self.reduced_layout:
            if x_label is not None:
                plt.getAxis('bottom').setLabel(text=x_label)
            if y_label is not None:
                plt.getAxis('left').setLabel(text=y_label)

        if title is not None:
            plt.setLabels(title=title)
        return plt

    def Add_Image_Item(self, return_plot=False, sidebar=False, stretch=1, title=None, tooltip_message=None):
        canvas = pg.GraphicsLayoutWidget()

        if tooltip_message is not None:
            canvas.ci.setToolTip(tooltip_message)

        #canvas.ci.layout.setContentsMargins(0, 0, 0, 0)
        #canvas.ci.layout.setSpacing(0)

        canvas.setBackground((255, 255, 255))

        self.Add_element(canvas, sidebar, stretch)

        plot = canvas.addPlot(row=0, col=0)
        #plot.setContentsMargins(0,0,0,0)

        if title is not None:
            plot.setLabels(title=title)

        plot.hideAxis('left')
        plot.hideAxis('bottom')
        image_item = pg.ImageItem(np.random.rand(291, 291, 3))

        plot.addItem(image_item)
        if return_plot:
            return image_item, plot
        else:
            return image_item


    def Add_plot_curve(self, title=None, return_plot=False, sidebar=False, stretch=1, number_of_curves=1, colors=[(0, 0, 0),(255, 0, 0),(0, 0, 255),(0, 250, 150),(255, 0, 255)], lines=[], names=[''], legend=False, x_label=None, y_label=None, return_list=False, always_show_axis_labels=False, tooltip_message=None):
        plt = self.Add_plot(title, sidebar, stretch, tooltip_message=tooltip_message)

        if always_show_axis_labels or not self.reduced_layout:
            if x_label is not None:
                plt.getAxis('bottom').setLabel(text=x_label)
            if y_label is not None:
                plt.getAxis('left').setLabel(text=y_label)

        if legend:
            plt.addLegend()

        for line in lines:
            plt.addLine(y=line,pen=(50,50,50,255))

        curves=[]
        for i in range(number_of_curves):
            curve = pg.PlotCurveItem([], name=names[i%len(names)], pen=colors[i%len(colors)])

            plt.addItem(curve)
            curves.append(curve)
        if number_of_curves == 1 and not return_list:
            curves = curves[0]

        if return_plot:
            return curves, plt
        else:
            return curves

    def Next_Tab(self, title='', stretch=1 , create_h_block=True):

        main_tab_container = QWidget()

        #def to_window(event):
        #    indx=self.tabs.indexOf(main_tab_container)
        #    self.tabs.removeTab(indx)
        #    main_tab_container.setParent(None)
        #    main_tab_container.show()

        def to_tab(event):
            self.tabs.addTab(main_tab_container, title)

        main_tab_container.closeEvent = to_tab
        main_tab_container.setWindowTitle(title)

        #bl = QHBoxLayout(self.main_tab_container)
        #bl.addWidget(self.net_info_text_label)
        self.tabs.addTab(main_tab_container, title)
        self.visualization_layout = QVBoxLayout(main_tab_container)
        if create_h_block:
            self.Next_H_Block(stretch)
        return main_tab_container


    def Next_H_Block(self,stretch=1):
        self.current_H_block = QHBoxLayout()
        self.visualization_layout.addLayout(self.current_H_block, stretch=stretch)

    def init_QT_Window(self, label, create_sidebar=True):
        self.width = 1200
        self.height = 600
        self.main_window.setWindowTitle(label)
        self.main_window.setGeometry(10, 40, self.width, self.height)


        self.main_h_layout = QHBoxLayout(self.main_window)

        self.main_split = QSplitter()#self.main_window
        self.main_h_layout.addWidget(self.main_split)

        self.sidebar_column_layout = QHBoxLayout()
        self.sidebar_widget=QWidget()
        self.sidebar_widget.setLayout(self.sidebar_column_layout)
        self.main_split.addWidget(self.sidebar_widget)
        #self.main_h_layout.addLayout(self.sidebar_column_layout, stretch=2)

        if create_sidebar:
            self.Add_New_Sidebar_Column()


        #cont = QVBoxLayout()

        self.vsplit = QSplitter()
        self.vsplit.setOrientation(Qt.Vertical)

        self.hsplit1=QSplitter()
        self.tabs = MyTabWidget(new=True)
        self.tabs2 = MyTabWidget(new=True)
        self.hsplit1.addWidget(self.tabs)
        self.hsplit1.addWidget(self.tabs2)
        self.hsplit1.setSizes([1,0])

        self.hsplit2=QSplitter()
        self.tabs3 = MyTabWidget(new=True)
        self.tabs4 = MyTabWidget(new=True)
        self.hsplit2.addWidget(self.tabs3)
        self.hsplit2.addWidget(self.tabs4)
        self.hsplit2.setSizes([1, 0])

        self.vsplit.addWidget(self.hsplit1)
        self.vsplit.addWidget(self.hsplit2)
        self.vsplit.setSizes([1, 0])

        #cont.addWidget(vsplit)

        self.main_split.addWidget(self.vsplit)
        self.main_split.setSizes([1, 18])
        #self.main_h_layout.addLayout(cont, stretch=8)


        #cont = QVBoxLayout()

        #self.tabs2.addTab(QWidget(),'test')

        #self.main_h_layout.addLayout(cont, stretch=8)

        #self.main_tab_container = QWidget()
        #self.visualization_layout = QVBoxLayout(self.main_tab_container)
        #self.tabs.addTab(self.main_tab_container, "Main")

        #self.Next_Tab('Main')

        self.tabs.currentChanged.connect(self._onTabChange)
        self.tabs2.currentChanged.connect(self._onTabChange)
        self.tabs3.currentChanged.connect(self._onTabChange)
        self.tabs4.currentChanged.connect(self._onTabChange)

    def _onTabChange(self, i):
        self.on_tab_change(i)

    def on_tab_change(self, i):
        return #overwrite

    def create_detail_label_pixmap(self):
        self.detail_qlabel = QLabel()
        self.detail_qlabel.setScaledContents(True)
        self.Add_Sidebar_Element(self.detail_qlabel,stretch=10)
        self.update_detail_image()

    def update_detail_image(self):
        self.detail_pixmap = QPixmap(self.detail_qlabel.width(), self.detail_qlabel.height())
        painter = QPainter(self.detail_pixmap)
        painter.eraseRect(0, 0, self.detail_qlabel.width(), self.detail_qlabel.height())
        self.refresh_detail_image()

    def refresh_detail_image(self):
        self.detail_qlabel.setPixmap(self.detail_pixmap)
        self.detail_qlabel.repaint()



    def create_content_label_pixmap(self):
        self.content_qlabel = QLabel()
        self.content_qlabel.setScaledContents(True)
        self.visualization_layout.addWidget(self.content_qlabel, stretch=1)
        self.update_content_image()

    def update_content_image(self):
        self.content_pixmap = QPixmap(self.content_qlabel.width(), self.content_qlabel.height())
        painter = QPainter(self.content_pixmap)
        painter.eraseRect(0, 0, self.content_qlabel.width(), self.content_qlabel.height())
        self.refresh_content_image()

    def refresh_content_image(self):
        self.content_qlabel.setPixmap(self.content_pixmap)
        self.content_qlabel.repaint()




    def Add_New_Sidebar_Column(self):
        self.sidebar_current_vertical_layout = QVBoxLayout()
        #self.sidebar_current_vertical_layout.setSpacing(0)
        #self.sidebar_current_vertical_layout.setContentsMargins(0, 0, 0, 0)
        self.sidebar_column_layout.addLayout(self.sidebar_current_vertical_layout, stretch=1)

    def Add_Sidebar_Spacing(self):
        self.Add_Sidebar_Element(QLabel())

    def Add_Sidebar_Element(self, elements=[], percentages=None, stretch=1, return_h_layout=False):
        if type(elements) is not list: elements=[elements]
        if percentages is None: percentages=[1 for _ in elements]

        self.sidebar_hblock = QHBoxLayout()
        #block.setSpacing(0)
        #block.setMargin(0)
        #block.setContentsMargins(0, 0, 0, 0)

        for element,p in zip(elements,percentages):
            self.sidebar_hblock.addWidget(element,stretch=p)

        self.sidebar_current_vertical_layout.addLayout(self.sidebar_hblock, stretch=stretch)
        if return_h_layout:
            return self.sidebar_hblock
        else:
            return elements

    def keyReleaseEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.control_key_down=False

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Control:
            self.control_key_down=True

        if event.key() in [Qt.Key_N, Qt.Key_C, Qt.Key_T, Qt.Key_M]:
            self.text_input_dialog('Please enter Note:', 'set note', self.set_note, default_text='-')

        if event.key() == Qt.Key_W:
            indx = self.tabs.currentIndex()
            if indx >= 0:
                widget=self.tabs.currentWidget()
                self.tabs.removeTab(indx)
                widget.setParent(None)
                widget.setWindowTitle('Warning: experimental feature (closing/reattaching the window can lead to crashes)')
                widget.show()

        if event.key() == Qt.Key_Space:
            self.pause = not self.pause

        if event.key() == Qt.Key_1:
            self.vsplit.setSizes([100000, 0])
            self.hsplit1.setSizes([100000, 0])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_2:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 0])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_3:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_4:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 100000])

        if event.key() == Qt.Key_5:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 0])
            self.hsplit2.setSizes([100000, 100000])

        if event.key() == Qt.Key_6:
            self.vsplit.setSizes([100000, 100000])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 0])

        if event.key() == Qt.Key_7:
            self.vsplit.setSizes([100000, 0])
            self.hsplit1.setSizes([100000, 100000])
            self.hsplit2.setSizes([100000, 0])

    def set_note(self, note_str):
        self.main_window.setWindowTitle(self.main_window.windowTitle() + ' | ' + note_str)
        if note_str=='-':
            self.main_window.setWindowTitle(self.main_window.windowTitle().split('|')[0])

    def text_input_dialog(self, txt, btn_text, func, default_text=''):
        dlg = QDialog()
        dlg.setWindowTitle(txt)
        layout = QVBoxLayout()

        input_le = QLineEdit(default_text)
        layout.addWidget(input_le)

        def btn_clicked():
            func(input_le.text())
            dlg.close()

        btn = QPushButton(btn_text)
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(300, 50)
        dlg.exec()


    def draw_qimage(self, qimg, x, y, w, h, pixmap=None):
        if pixmap is None:
            pixmap=self.content_pixmap
        painter = QPainter(pixmap)
        painter.drawImage(QtCore.QRect(x, y, w, h), qimg, QtCore.QRect(0, 0, qimg.width(), qimg.height()))


    def numpy_array_to_qimage(self, data, byte_count=3, alpha=255):
        data = data.astype(int)
        bytesPerLine = byte_count * data.shape[1]

        if byte_count == 1: format = QImage.Format_Grayscale8
        if byte_count == 3: format = QImage.Format_RGB888
        if byte_count == 4: format = QImage.Format_RGBA8888

        if byte_count == 4 and alpha < 255:
            data[:, :, 3] = alpha

        return QImage(np.require(data, np.uint8, 'C'), data.shape[1], data.shape[0], bytesPerLine, format)




from PyQt5.QtWidgets import QTabWidget
from PyQt5.QtCore import Qt, QPoint, QMimeData
from PyQt5.QtGui import QPixmap, QRegion, QDrag, QCursor

class MyTabWidget(QTabWidget):

   def __init__(self, parent=None, new=None):
      super().__init__(parent)
      self.setAcceptDrops(True)
      self.tabBar().setMouseTracking(True)
      self.setMovable(True)
      if new:
         MyTabWidget.setup(self)

   def __setstate__(self, data):
      self.__init__(new=False)
      self.setParent(data['parent'])
      for widget, tabname in data['tabs']:
         self.addTab(widget, tabname)
      MyTabWidget.setup(self)

   def __getstate__(self):
      data = {
         'parent' : self.parent(),
         'tabs' : [],
      }
      tab_list = data['tabs']
      for k in range(self.count()):
         tab_name = self.tabText(k)
         widget = self.widget(k)
         tab_list.append((widget, tab_name))
      return data

   def setup(self):
       pass

   def mouseMoveEvent(self, e):
       if e.buttons() != Qt.RightButton:
           return

       globalPos = self.mapToGlobal(e.pos())
       tabBar = self.tabBar()
       posInTab = tabBar.mapFromGlobal(globalPos)
       index = tabBar.tabAt(e.pos())
       tabBar.dragged_content = self.widget(index)
       tabBar.dragged_tabname = self.tabText(index)
       tabRect = tabBar.tabRect(index)

       pixmap = QPixmap(tabRect.size())
       tabBar.render(pixmap, QPoint(), QRegion(tabRect))
       mimeData = QMimeData()

       drag = QDrag(tabBar)
       drag.setMimeData(mimeData)
       drag.setPixmap(pixmap)

       cursor = QCursor(Qt.OpenHandCursor)

       drag.setHotSpot(e.pos() - posInTab)
       drag.setDragCursor(cursor.pixmap(), Qt.MoveAction)
       drag.exec_(Qt.MoveAction)

   def dragEnterEvent(self, e):
       e.accept()
       # self.parent().dragged_index = self.indexOf(self.widget(self.dragged_index))

   def dragLeaveEvent(self, e):
       e.accept()

   def dropEvent(self, e):
       if e.source().parentWidget() == self:
           return

       e.setDropAction(Qt.MoveAction)
       e.accept()
       tabBar = e.source()
       self.addTab(tabBar.dragged_content, tabBar.dragged_tabname)

#tqt = TREN_QT_UI_Base(None)
#tqt.show()

#########################
#########################Manual
#########################

'''
self.create_detail_label_pixmap()
self.create_content_label_pixmap()

for i in range(5):
    self.Add_Sidebar_Element([QPushButton('refresh', self.main_window), QPushButton('refresh', self.main_window)])


def onclick(event):
    self.update_detail_image()
    self.update_content_image()

    painter = QPainter(self.detail_pixmap)
    pen = QPen(QColor(255, 255, 0), 3)
    painter.setPen(pen)
    painter.drawLine(10, 10, 100, 100)
    painter.drawRect(painter.clipBoundingRect())

    self.refresh_content_image()
    self.refresh_detail_image()

    # print('adssgdf')


for i in range(5):
    self.Add_Sidebar_Element(QPushButton('refresh', self.main_window))[0].clicked.connect(onclick)

self.Add_Sidebar_Spacing()

self.Add_New_Sidebar_Column()

self.Add_Sidebar_Spacing()

for i in range(5):
    self.Add_Sidebar_Element(QPushButton('refresh', self.main_window))
'''




from PymoNNto.NetworkCore.Network import *
import numpy as np
import matplotlib.pylab as plt
from collections import Counter
import scipy.sparse as sp


class Reconstruct_Analyze_Label_Network():
    def __init__(self, network):
        self.network = network


    def get_network_responses(self, pattern_group, target_activation_neurons, target_recording_neuron_groups, max_gamma_cycles_for_pattern = 2, pattern_repeat=1, shuffle=True, combinations=False):

        activators={}
        for ng in self.network.NeuronGroups:
            ng.learning=False
            for be in ng.behaviour.values():
                if isinstance(be, NeuronActivator) and not isinstance(be, NeuronManualActivator):
                    activators[be] = be.behaviour_enabled
                    be.behaviour_enabled = False


        man_act = None

        for be in target_activation_neurons.behaviour.values():
            if isinstance(be, NeuronManualActivator):
                man_act = be

        if man_act is None:
            man_act=self.network.add_behaviours_to_neuron_group([NeuronManualActivator()], target_activation_neurons)[0]

        if target_recording_neuron_groups is None:
            target_recording_neuron_groups = self.network.NeuronGroups

        activity_trace = []
        pattern_ids = []

        pattern_steps = []

        pgp=pattern_group.patterns
        if len(pattern_group.patterns) == 0:
            pgp=[pattern_group.get_pattern() for _ in range(100)]

        for _ in range(pattern_repeat):

            for i in range(len(pgp)):

                if combinations:
                    for i2 in range(len(pgp)):
                        pattern_steps.append([i, i2])
                else:
                    pattern_steps.append([i])

        if shuffle:
            random.shuffle(pattern_steps)

        pattern_steps += [-1 for _ in range(max_gamma_cycles_for_pattern)]

        for i in pattern_steps:
            if i != -1:
                id=''
                for p in i:
                    pattern = pgp[p]
                    man_act.activate(pattern)
                    id+='{}'.format(pattern)

                pattern_ids.append(id)
            #else:
            #    pattern_ids.append('[]')

            self.network.simulate_iteration()

            activation_step = []
            for neuron_group in target_recording_neuron_groups:
                activation_step.append(neuron_group.activity.copy())#.output_activity_history[0].copy())
            activity_trace.append(np.array(activation_step).flatten())


        for ng in self.network.NeuronGroups:
            ng.learning = True

        for be in activators:
            be.behaviour_enabled = activators[be]

        return np.array(activity_trace), np.array(pattern_ids)


    def get_random_network_responses(self, iterations, neuron_groups=None, PatternGroups=None):
        activity_trace = []
        pattern_ids = []

        if neuron_groups is None:
            neuron_groups = self.network.NeuronGroups

        for ng in neuron_groups: ng.learning = False
        for pg in PatternGroups:
            pg.individual_possibility = 0.0
            pg.group_possibility = 1.0
        #LGN_PC_Neurons[TRENNeuronActivator].TNAPatterns[0].individual_possibility = 0.0
        #LGN_PC_Neurons[TRENNeuronActivator].TNAPatterns[0].group_possibility = 1.0

        for i in range(iterations):
            self.network.simulate_iteration()

            activation_step = []
            for neuron_group in neuron_groups:
                activation_step.append(neuron_group.activity.copy())#.output_activity_history[0].copy())
            activity_trace.append(np.array(activation_step).flatten())

            if PatternGroups is not None:
                pattern_step = ""
                for patter_group in PatternGroups:
                    pattern_step += "{}".format(patter_group.current_pattern_index)
                pattern_ids.append(pattern_step)

        for ng in neuron_groups: ng.learning = True
        #for pg in PatternGroups:
        #    pg.individual_possibility = 0.0
        #    pg.group_possibility = 1.0

        return np.array(activity_trace), np.array(pattern_ids)


    def single_reconstruction(self, dest_group, synapse_weight_attribute_name, reconstruction_steps, individual_norm=False, print_status=False):
        timestep_reconstructions = []

        for step in range(reconstruction_steps):
            if print_status:
                print('reconstruction', step)

            for neurons in self.network.NeuronGroups:
                neurons.reconstruction_act_temp *= 0

            for syn in self.network.SynapseGroups:
                if hasattr(syn, synapse_weight_attribute_name):
                    syn.src.reconstruction_act_temp += np.dot(getattr(syn, synapse_weight_attribute_name).transpose(),
                                                              syn.dst.reconstruction_act)  # [][syn.dst.mask]

            for neurons in self.network.NeuronGroups:
                neurons.reconstruction_act = neurons.reconstruction_act_temp.copy()  # +=

            timestep_reconstructions.append(dest_group.reconstruction_act.copy())

        timestep_reconstructions = np.array(timestep_reconstructions)
        if individual_norm:
            timestep_reconstructions /= np.max(timestep_reconstructions)

        return timestep_reconstructions


    def get_network_activity_reconstruction(self, dest_group, synapse_weight_attribute_name, recostruction_steps, individual_norm=False, print_status=False, active_groups=None):

        for neurons in self.network.NeuronGroups:
            neurons.reconstruction_act = neurons.output_activity_history[0].copy()
            if active_groups is not None and not neurons in active_groups:
                neurons.reconstruction_act *= 0
            neurons.reconstruction_act_temp = np.zeros(neurons.size)

        return self.single_reconstruction(dest_group, synapse_weight_attribute_name, recostruction_steps, individual_norm, print_status=False)


    def get_reconstruction(self, source_group, source_group_neuron_indexes, dest_group, synapse_weight_attribute_name, recostruction_steps, individual_norm=False, print_status=False, single_timestep=True):

        if source_group_neuron_indexes is None:
            source_group_neuron_indexes = list(range(source_group.size))

        for neurons in self.network.NeuronGroups:
            neurons.reconstruction_act = np.zeros(neurons.size)
            neurons.reconstruction_act_temp = np.zeros(neurons.size)

        result = []

        for neuron_index in source_group_neuron_indexes:
            if print_status:
                print('neuron', neuron_index, 'reconstruction')

            for neurons in self.network.NeuronGroups:
                neurons.reconstruction_act *= 0

            source_group.reconstruction_act[neuron_index] = 1

            timestep_reconstructions = self.single_reconstruction(dest_group, synapse_weight_attribute_name, recostruction_steps, individual_norm, print_status=False)

            if single_timestep:
                result.append(np.sum(timestep_reconstructions, axis=0))
            else:
                result.append(timestep_reconstructions)

        result = np.array(result)

        #if individual_norm:#todo test
        #    result = result / np.max(result, axis=1)[:, None]

        return np.array(result)

    def contains_with_gap(self, string, substring, gapchar='|'):
        result = 0
        text = []
        for s in range(len(string) - len(substring)):
            match = True
            for ss in range(len(substring)):
                if string[s + ss] != substring[ss] and (substring[ss] != gapchar):
                    match = False
            if match:
                result += 1
                if len(substring)>1:#exclude default characters
                    text.append(string[s - 5:s + len(substring) + 5])
        #print(set(text))
        return result, np.unique(text)#list(set(text))






    def ng_list(self, ng_list_or_name):
        if ng_list_or_name is None:
            return []
        elif type(ng_list_or_name)==list:
            return ng_list_or_name
        elif ng_list_or_name=='all':
            return self.network.NeuronGroups
        else:
            return self.network.getNG(ng_list_or_name)



    def zero_recon(self):
        for ng in self.network.NeuronGroups:
            ng.recon = ng.get_neuron_vec()


    #propagation variables:
    #- ng.recon             (needs to be set before call)
    #- ng.recon_temp        (temporal var)
    #- ng.layer_recon       (optional)
    #- ng.temporal_recon    (optional)

    #weight_attribute_name  name of synapse attribute
    #recon_steps            (steps of recon)

    #propagation_mode:
        # forward
        # backward

    #compute_layers: Bool   (written into ng.layer)
    #clip_min               (min recon values)
    #clip_max               (max recon values)

    #accumulation_mode:
        # addititve         recon+=new_recon
        # first_touch       recon=new_recon if recon==0
        # forget            recon=new_recon

    #synapse_filter:
        # 'all'             all
        # 'biggest'         biggest
        # [float]           syn >= max_syn*[float]

    #temporal_recon_groups  (optional)
        #[]                 disabled
        #[ng1,ng2,...]      temporal recon for ng1 and ng2

    #layer_compute_groups  (optional)
        #[]                 disabled
        #[ng1,ng2,...]      layer computation for ng1 and ng2

    #todo inhibitory and excitatory synapses
    #todo activation function (default:f(x)=x)



    def propagation(self, weight_attribute_name,  recon_steps, propagation_mode='forward', accumulation_mode='addititve', synapse_filter='all', clip_min=None, clip_max=None, temporal_recon_groups=[], layer_compute_groups=[], exponent=None, normalize=None, filter_weakest_percent=None):

        for ng in self.ng_list(layer_compute_groups):
            ng.layer_recon = ng.get_neuron_vec()+np.inf

        for ng in self.ng_list(temporal_recon_groups):
            ng.temporal_recon = []

        for step in range(recon_steps):

            for ng in self.network.NeuronGroups:
                ng.recon_temp = ng.get_neuron_vec()


            if propagation_mode == 'simulation':#todo implement
                for ng in self.network.NeuronGroups:#feed in
                    ng.activity = ng.recon.copy()
                self.network.learning_off()
                self.network.simulate_iteration()#simulate
                self.network.learning_on()
                for ng in self.network.NeuronGroups:#read out
                    ng.recon_temp = ng.activity
            else:
                for syn in self.network.SynapseGroups:
                    if hasattr(syn, weight_attribute_name) and 'GLU' in syn.tags:

                        w = getattr(syn, weight_attribute_name)
                        if type(w) == sp.csc.csc_matrix:
                            w = np.array(w.todense())

                        if synapse_filter == 'all':
                            w_mul = True
                        if synapse_filter == 'biggest':
                            # w_mul=w==w.max(axis=1)
                            w_mul = np.equal(w, np.tile(w.max(axis=1), (w.shape[1], 1)).transpose())  # better way!!!
                        if type(synapse_filter) == float:
                            # w_mul = w >= (w.max(axis=1)*synapse_filter)
                            w_mul = np.greater(w, np.tile(w.max(axis=1) * synapse_filter, (w.shape[1], 1)))

                        if propagation_mode == 'backward':
                            syn.src.recon_temp += (w * w_mul).transpose().dot(syn.dst.recon)  # disabled???

                        if propagation_mode == 'forward':
                            syn.dst.recon_temp += (w * w_mul).dot(syn.src.recon)  # disabled???

            if exponent is not None:
                syn.dst.recon_temp = np.power(syn.dst.recon_temp, exponent)

            if filter_weakest_percent is not None:
                recon_min = np.min(syn.dst.recon_temp)
                recon_max = np.max(syn.dst.recon_temp)
                syn.dst.recon_temp *= syn.dst.recon_temp>=(recon_min+(recon_max-recon_min)/100.0*filter_weakest_percent)

            if normalize is not None:
                s=np.sum(np.abs(syn.dst.recon_temp))
                if s>0:
                    syn.dst.recon_temp/=s




            for ng in self.ng_list(layer_compute_groups):
                ng.layer_recon[(ng.recon_temp != 0) * (ng.layer_recon == np.inf)] = step + 1

            for ng in self.network.NeuronGroups:

                if accumulation_mode == 'additive':
                    ng.recon += ng.recon_temp#.copy()??? not neccessary because of ng.get_neuron_vec() ?

                if accumulation_mode == 'first_touch':
                    ng.recon += ng.recon_temp*(ng.recon==0)

                if accumulation_mode == 'forget':
                    ng.recon = ng.recon_temp

                if clip_min is not None or clip_max is not None:
                    ng.recon = np.clip(ng.recon, clip_min, clip_max)



            for ng in self.ng_list(temporal_recon_groups):
                ng.temporal_recon.append(ng.recon.copy())





    #dict([(char,source.char_index_to_one_hot()) for char in range(source.A)])
    def label_and_group_neurons(self, label_ng, label_vectors, weight_attribute_name, steps):

        for ng in self.network.NeuronGroups:
            ng.class_label = []
            ng.temporal_layer = ng.get_neuron_vec()+np.inf


        for label_vec in label_vectors:
            self.zero_recon()
            label_ng.recon = label_vec.copy()

            label_ng.temporal_layer[label_vec != 0] = 0 #set first layer

            self.propagation(recon_steps=steps, propagation_mode='forward', accumulation_mode='first_touch', synapse_filter='biggest', clip_min=None, clip_max=None, temporal_recon_groups=[], layer_compute_groups='all', weight_attribute_name=weight_attribute_name)

            for ng in self.network.NeuronGroups:
                ng.class_label.append(ng.recon.copy())
                ng.temporal_layer = np.minimum(ng.temporal_layer, ng.layer_recon)

        for ng in self.network.NeuronGroups:
            ng.class_label = np.argmax(np.array(ng.class_label), axis=0)


    def synapse_iterate(self, func, x_attribute_name, y_attribute_name, weight_attribute_name, groups='all', weight_limit=1, include_fist_layer=False, n_biggest=None):

        #inf!!!

        for syn in self.network.SynapseGroups:
            if syn.src in self.ng_list(groups) and syn.dst in self.ng_list(groups):

                sg_indx = self.ng_list(groups).index(syn.src)
                dg_indx = self.ng_list(groups).index(syn.dst)

                #if group_shift:
                #    d_ng_y = self.ng_list(groups).index(syn.dst) * ng.size#* h * max_class_label * group_shift
                #    s_ng_y = self.ng_list(groups).index(syn.src) * ng.size#* h * max_class_label * group_shift
                #else:
                #    d_ng_y = 0
                #    s_ng_y = 0

                W = getattr(syn, weight_attribute_name)
                if type(W) == sp.csc.csc_matrix:
                    W = np.array(W.todense())
                w_global_max = np.max(W)

                src_x = getattr(syn.src, x_attribute_name)
                src_y = getattr(syn.src, y_attribute_name)
                dst_x = getattr(syn.dst, x_attribute_name)
                dst_y = getattr(syn.dst, y_attribute_name)
                for d in range(syn.dst.size):
                    print('\r', d, '/', syn.dst.size, end='')
                    wmax = np.max(W[d])
                    sorted_w = W[d].copy()
                    sorted_w.sort()
                    if wmax != 0:
                        for s in range(syn.src.size):
                            w = W[d, s]
                            if w != 0 and (dst_x[d] != 0 or include_fist_layer) and (weight_limit is None or w >= wmax * weight_limit) and (n_biggest is None or w==sorted_w[-n_biggest]):# and w >= wmax * weight_limit:# and w==wmax:# and w > wmax * 0.3333: #and w==wmax
                                func(s, src_x[s], src_y[s], d, dst_x[d], dst_y[d], w, sorted_w, w_global_max, sg_indx, dg_indx)



    #def multiple_timescale_layer_adjustment(self, groups='all'):

    #    for ng in self.ng_list(groups):

    #        if

    #def label

    def visualize_label_and_group_neurons(self, x_attribute_name, y_attribute_name, weight_attribute_name, groups='all', weight_limit=0.3333, group_shift=0, source=None, include_fist_layer=False, n_biggest=None, return_graph=False):
        max_layers = int(np.max([np.max(ng.temporal_layer[ng.temporal_layer!=np.inf]) for ng in self.ng_list(groups)]))
        max_class_label = int(np.max([np.max(ng.class_label) for ng in self.ng_list(groups)]))

        for ng in self.ng_list(groups):
            ng.i = np.arange(ng.size)

        h = np.sum([ng.size for ng in self.ng_list(groups)])

        #plt.xlim(left=-1)
        plt.axhline(0, color='gray')
        for char in range(max_class_label+1):#*len(self.ng_list(groups))
            plt.axhline(h * (char + 1), color='gray')
            if source is not None:
                c = ''# source.corpus[0:100000].count(source.index_to_symbol(char))
                plt.text(max_layers, h * char + 25, '\'{}\'{}'.format(source.index_to_char(char), c))

        result_neurons = []
        result_synapses = []

        #draw synapses
        def loop_func(s, sx, sy, d, dx, dy, w, sorted_w, w_global_max, sg_indx, dg_indx):
            if sx != np.inf and dx != np.inf:
                w_norm = w / w_global_max  # wmax
                if sy == dy:
                    color = [0, 1, 0, w_norm]
                else:

                    if sx < dx:  # forward =>black
                        color = [0, 0, 0, w_norm]
                    else:  # backward or sideways =>red
                        color = [1, 0, 0, w_norm]
                        # color[2] = 1
                        # color[3] = w_norm# * 0.1

                if return_graph:
                    result_synapses.append([s, d])
                else:
                    plt.plot([dx, sx], [d + h * dy + dg_indx, s + h * sy + sg_indx], '-k', color=color)

        self.synapse_iterate(loop_func, x_attribute_name=x_attribute_name, y_attribute_name=y_attribute_name, weight_attribute_name=weight_attribute_name, groups=groups, weight_limit=weight_limit, include_fist_layer=include_fist_layer, n_biggest=n_biggest)

        #draw_neurons
        for ng in self.ng_list(groups):#todo group neuron_groups or filter
            ng_y = self.ng_list(groups).index(ng) * ng.size#* h * max_class_label * group_shift
            x_positions = getattr(ng, x_attribute_name)
            y_positions = getattr(ng, y_attribute_name)
            for x in np.unique(x_positions):#range(max_layers+1)
                ng_color = 'black'
                if hasattr(ng, 'color'):
                    ng_color = ng.color

                lmask = x_positions == x
                if x==np.inf:
                    ng_color = 'red'
                    x=-1
                if return_graph:
                    for i,l in enumerate(lmask):
                        if l:
                            result_neurons.append([x, np.arange(ng.size)[i] + h * y_positions[i]+ ng_y])
                else:
                    plt.plot(np.ones(np.sum(lmask)) * x, np.arange(ng.size)[lmask] + h * y_positions[lmask]+ ng_y, 'ok', ms=3, color=ng_color)


        if return_graph:
            return result_neurons, result_synapses
        else:
            plt.show()

    def get_neuron_positions(self, ng, x_attribute_name, y_attribute_name, y_scale=1.0):
        x_positions = np.array(getattr(ng, x_attribute_name))
        y_positions = np.array(getattr(ng, y_attribute_name))
        size=len(x_positions)
        return np.array(list(zip(x_positions+np.random.rand(size)*0.5-0.25, y_positions+np.random.rand(size)*0.5-0.25)))#(np.arange(ng.size)+np.arange(ng.size)*y_positions)*y_scale



    def get_all_buffer_length_distribution(self, groups):
        max_class_label = int(np.max([np.max(ng.class_label) for ng in self.ng_list(groups)]))
        max_temporal_layer = int(np.max([np.max(ng.temporal_layer[ng.temporal_layer!=np.inf]) for ng in self.ng_list(groups)]))

        max_label_buffers = []
        label_buffer_length_distribution=[]
        for label in range(max_class_label):

            buffer_elements=[]
            for ng in [self.network.NeuronGroups[0]]:
                buffer_elements+=list(ng.temporal_layer[(ng.class_label==label) * (ng.temporal_layer!=np.inf)])

            max_buffer_length=np.max(buffer_elements)

            max_label_buffers.append(max_buffer_length)

            buffer_length_distribution=[]
            for l in range(max_temporal_layer+1):#range(int(max_buffer_length+1)):
                occurences=buffer_elements.count(l)
                buffer_length_distribution.append(occurences)

            label_buffer_length_distribution.append(buffer_length_distribution)

        average_max=np.average(max_label_buffers)
        return max_label_buffers, label_buffer_length_distribution, average_max


    def get_synapse_transition_dict(self, source, x_attribute_name, y_attribute_name, weight_attribute_name, groups='all', weight_limit=0.3333, include_fist_layer=False, n_biggest=None):

        #forward_same = 0
        #forward_different = 0
        #backward = 0
        transitions = {}

        def inc(key):
            if not key in transitions:
                transitions[key] = 1
            else:
                transitions[key] += 1

        def loop_func(s, sx, sy, d, dx, dy, w, sorted_w, w_global_max, sg_indx, dg_indx):
            #if sy == dy:
            #    inc('forward_same')
            #else:
            #    inc('forward_different')
            #if sx >= dx:
            #    inc('backward')'

            ss = source.index_to_char(int(sy))
            ds = source.index_to_char(int(dy))

            key = 'UNKNOWN'

            if dx != np.inf and sx != np.inf:

                dist = int(dx - sx)

                if dist > 1:
                    key = ss + ''.join(['-' for _ in range(dist - 2)]) + ds

                if dist <= 0:
                    key = ss + ''.join(['|' for _ in range(-dist)]) + ds

                if ss == ds:  # buffer
                    if dist > 0:
                        key = ss  # dist does not matter
            # if len(key) == 2:
            #    print(key)
            inc(key)

        self.synapse_iterate(loop_func, x_attribute_name=x_attribute_name, y_attribute_name=y_attribute_name, weight_attribute_name=weight_attribute_name, groups=groups, weight_limit=weight_limit, include_fist_layer=include_fist_layer, n_biggest=n_biggest)
        return transitions

    def get_transition_frequencies(self, transitions, source):
        frequency_dict = {}
        occurence_dict = {}
        filtered_transitions = {}
        for key in transitions:
            found, occurences = self.contains_with_gap(source.corpus, key, gapchar='|')
            frequency_dict[key] = found
            occurence_dict[key] = occurences
            if found > 0:
                filtered_transitions[key] = transitions[key]

        return frequency_dict, occurence_dict, filtered_transitions


    def plot_trend_line(self, x, y):
        plt.plot(x, np.poly1d(np.polyfit(x, y, 1))(x))

        fit = np.polyfit(x, y, 1, cov=True)  # linear

        unc = np.sqrt(np.diag(fit[1]))

        slope = fit[0][0]
        intercept = fit[0][1]
        slopes = np.linspace(slope - unc[0], slope + unc[0], 20)  # take 20 slopes in unc range
        intercepts = np.linspace(intercept - unc[1], intercept + unc[1], 20)  # take 20 intercepts in unc range

        line_array = []
        for i in range(len(slopes)):
            line_array.append(x * slopes[i] + intercepts[i])
        line_array = np.array(line_array)
        lower_edge = np.min(line_array, axis=0)
        upper_edge = np.max(line_array, axis=0)

        # slopes = np.linspace(slope - unc[0], slope + unc[0], 20)  # take 20 slopes in unc range
        # intercepts = np.linspace(intercept - unc[1], intercept + unc[1], 20)
        plt.fill_between(x, lower_edge, upper_edge, color='gray', alpha=0.4)


    def plot_transition_frequencies(self, transitions, frequencies, plot_single=False, plot_multiple=True, show_labels=True):

        txt = []
        x = []
        y = []
        for key in transitions:
            if key != 'UNKNOWN' and ((plot_single and len(key)==1) or (plot_multiple and len(key)>1)):
                txt.append(key)
                x.append(transitions[key])
                y.append(frequencies[key])
        plt.scatter(x, y)

        if show_labels:
            for txt,xv,yv in zip(txt,x,y):
                plt.annotate(txt.replace(' ','_'), (xv, yv))

        return np.array(x).astype(def_dtype), np.array(y).astype(def_dtype)

    def get_transition_classes(self, transitions, frequencies):

        buffer = 0
        useful = 0
        useless = 0

        for key in transitions:
            if len(key) == 1:
                buffer += transitions[key]
            elif frequencies[key]>0:
                useful += transitions[key]
            else:
                useless += transitions[key]
            #unknown automatically counted to useless

        return {'BUFFER': buffer, 'USEFUL': useful, 'USELESS': useless}

        # print('useful:', useful_conn, 'forward_same', forward_same, 'forward_different', forward_different, 'backward', backward)

    def plot_bars(self, transitions):
        key_l = []
        value_l = []
        for key, value in sorted(transitions.items(), key=lambda item: item[1]):
            key_l.append(key)
            value_l.append(value)

        rects = plt.bar(np.arange(len(key_l)) * 10 - 4, value_l, width=8)
        for rect, label in zip(rects, key_l):
            height = rect.get_height()
            plt.text(rect.get_x() + rect.get_width() / 2, height, label.replace(' ', '_'), ha='center', va='bottom')



import threading
import sys
import time

def start_SM_Listener(storage_manager):

    def get_input():
        while True:
            data = input()
            print(data)

    global exit_flag
    exit_flag=False
    global input_thread
    input_thread = threading.Thread(target=get_input, args=(lambda : exit_flag, ))
    input_thread.start()

def terminate_listener():
    global exit_flag
    exit_flag=True

start_SM_Listener(None)

time.sleep(10)

terminate_listener()

